@@ -5,6 +5,7 @@ import re
 
 from genshi.builder import tag
 
+import trac
 from trac.config import ListOption, Option
 from trac.core import Component, implements
 from trac.util.translation import _
@@ -12,6 +13,7 @@ from trac.versioncontrol.api import is_default, NoSuchChangeset, RepositoryManag
 from trac.versioncontrol.web_ui.changeset import ChangesetModule
 from trac.web.api import IRequestHandler
 from trac.web.auth import LoginModule
+from trac.web.chrome import add_warning
 
 
 class GitHubLoginModule(LoginModule):
@@ -26,8 +28,18 @@ class GitHubLoginModule(LoginModule):
             # Use the same names as LoginModule to avoid duplicates.
             yield ('metanav', 'login', _('logged in as %(user)s',
                                          user=req.authname))
-            yield ('metanav', 'logout',
-                   tag.a(_('Logout'), href=req.href.github('logout')))
+            from pkg_resources import parse_version
+            if parse_version(trac.__version__) < parse_version('1.0.2'):
+                yield ('metanav', 'logout',
+                       tag.a(_('Logout'), href=req.href.github('logout')))
+            else:
+                yield ('metanav', 'logout',
+                       tag.form(tag.div(tag.button(_('Logout'),
+                                                   name='logout',
+                                                   type='submit')),
+                                action=req.href.github('logout'),
+                                method='post', id='logout',
+                                class_='trac-logout'))
         else:
             # Use a different name from LoginModule to allow both in parallel.
             yield ('metanav', 'github_login',
@@ -57,13 +69,24 @@ class GitHubLoginModule(LoginModule):
         req.redirect(authorization_url)
 
     def _do_oauth(self, req):
-        oauth = self._oauth_session(req)
+        try:
+            state = req.session['oauth_state']
+        except KeyError as exc:
+            self._reject_oauth(req, exc)
+
+        oauth = self._oauth_session(req, state)
+
         authorization_response = req.abs_href(req.path_info) + '?' + req.query_string
         client_secret = self._client_config('secret')
-        oauth.fetch_token(
-            'https://github.com/login/oauth/access_token',
-            authorization_response=authorization_response,
-            client_secret=client_secret)
+        # Inner import to avoid a hard dependency on requests-oauthlib.
+        import oauthlib
+        try:
+            oauth.fetch_token(
+                'https://github.com/login/oauth/access_token',
+                authorization_response=authorization_response,
+                client_secret=client_secret)
+        except oauthlib.oauth2.OAuth2Error as exc:
+            self._reject_oauth(req, exc)
 
         user = oauth.get('https://api.github.com/user').json()
         # Small hack to pass the username to _do_login.
@@ -74,12 +97,26 @@ class GitHubLoginModule(LoginModule):
 
         return super(GitHubLoginModule, self)._do_login(req)
 
-    def _oauth_session(self, req):
+    def _reject_oauth(self, req, exc):
+            self.log.warn(exc)
+            add_warning(req, _("Invalid request. Please try to login again."))
+            self._redirect_back(req)
+
+    def _do_logout(self, req):
+        req.session.pop('oauth_state', None)
+        super(GitHubLoginModule, self)._do_logout(req)
+
+    def _oauth_session(self, req, state=None):
         client_id = self._client_config('id')
         redirect_uri = req.abs_href.github('oauth')
         # Inner import to avoid a hard dependency on requests-oauthlib.
         from requests_oauthlib import OAuth2Session
-        return OAuth2Session(client_id, redirect_uri=redirect_uri, scope=[])
+        return OAuth2Session(
+            client_id,
+            scope=[''],
+            redirect_uri=redirect_uri,
+            state=state,
+        )
 
     def _client_config(self, key):
         assert key in ('id', 'secret')
@@ -90,7 +127,7 @@ class GitHubLoginModule(LoginModule):
             return os.environ.get(value, '')
         else:
             with open(value) as f:
-                return f.read.strip()
+                return f.read().strip()
 
 
 class GitHubMixin(object):
@@ -171,6 +208,16 @@ class GitHubPostCommitHook(GitHubMixin, Component):
             return True
 
     def process_request(self, req):
+        path = req.args['path']
+
+        rm = RepositoryManager(self.env)
+        reponame, repos, path = rm.get_repository_by_path(path)
+
+        if repos is None or path != '/':
+            msg = u'No such repository (%s)\n' % path
+            self.log.warning(msg.rstrip('\n'))
+            req.send(msg.encode('utf-8'), 'text/plain', 400)
+
         if req.method != 'POST':
             msg = u'Endpoint is ready to accept GitHub notifications.\n'
             self.log.warning(u'Method not allowed (%s)' % req.method)
@@ -185,16 +232,6 @@ class GitHubPostCommitHook(GitHubMixin, Component):
             self.log.warning(msg.rstrip('\n'))
             req.send(msg.encode('utf-8'), 'text/plain', 400)
 
-        path = req.args['path']
-
-        rm = RepositoryManager(self.env)
-        reponame, repos, path = rm.get_repository_by_path(path)
-
-        if path != '/':
-            msg = u'No such repository (%s)\n' % path
-            self.log.warning(msg.rstrip('\n'))
-            req.send(msg.encode('utf-8'), 'text/plain', 400)
-
         output = u'Running hook on %s\n' % (reponame or '(default)')
 
         output += u'* Updating clone\n'
